# Abaqus read plugin .inp mesh

namespace eval Abaqus {       
    variable info_plugin ;#information read from .xml plugin file    
    #internal use variables
    variable percent_done
    variable stop
    variable _merge  0 ;#to set in file browser options if want to merge or delete old model+results
    variable debug 0

    set info_plugin(path) [file dirname [info script]] ;#to know where to find the files
    if { [info exists info_plugin(Version)] } {
        set other_plugin_version $info_plugin(Version)
    }
    array set info_plugin [ReadProblemtypeXml [file join $info_plugin(path) [file tail $info_plugin(path)].xml] InfoPlugin {Name Version MinimumGiDVersion}]    
    if { [info exists other_plugin_version] && $other_plugin_version != $info_plugin(Version)} {
        WarnWinText [_ "There are two versions of %s: the %s and %s inside the plugins folder. One must be deleted" $info_plugin(Name) $other_plugin_version $info_plugin(Version)]
    }    
    if { [GidUtils::VersionCmp $info_plugin(MinimumGiDVersion)] < 0 } {  
        WarnWinText [_ "This plugin requires GiD %s or later" $info_plugin(MinimumGiDVersion)]
    }
}

#to add more options to the standar file browser
proc Abaqus::MoreImportOptions { f } {
    variable _merge 
    ttk::checkbutton $f.chkGeometry -text [_ "Merge"] -variable Abaqus::_merge   
    grid columnconfigure $f 1 -weight 1
    grid $f.chkGeometry -sticky w -columnspan 2  

    variable export_dir
    if { ![info exists export_dir] } {
        #trick to show the more options
        set export_dir open
    }
    return Abaqus::export_dir
}

proc Abaqus::ReadPreW { } {
    variable _merge 
    set multiple 1
    set filenames [MessageBoxGetFilename file read [_ "Read Abaqus mesh"] \
                       {} [list [list {Abaqus} {.inp }] [list [_ "All files"] {.*}]] {} $multiple \
            [list [_ "Import options"] Abaqus::MoreImportOptions]]
    if { [llength $filenames] } {
        if { !$_merge } {
            GiD_Process Mescape Files New No escape ;#to avoid previous geometry
        }
        GidUtils::EvalAndEnterInBatch Abaqus::ReadPre $filenames
    }    
}


#to drag and drop a file
proc Abaqus::ReadPreSingleFile { filename } {
    if { !$_merge } {
        GiD_Process Mescape Files New No escape ;#to avoid previous geometry
    }
    GidUtils::EvalAndEnterInBatch Abaqus::ReadPre [list $filename]
}

proc Abaqus::ReadPre { filenames } {
    set fail 0
    GidUtils::WaitState .gid
    GidUtils::DisableGraphics               
    if { [catch { set fail [Abaqus::ReadPre0 $filenames] } err options] } {
        set fail 1
    }
    GidUtils::EnableGraphics
    GidUtils::EndWaitState .gid
    if { $fail } {
        return -options $options $err
    } else {
        MeshView 0        
        GiD_Process 'Zoom Frame escape
        GidUtils::UpdateWindow LAYER
        GidUtils::UpdateWindow GROUPS
        GidUtils::SetWarnLine [_ "Geometry read"]
    }    
    return $fail
}

proc Abaqus::GetGiDElementTypeAndNnodes { abaqus_element_type } {
    variable debug
    set linear_2 {B21 B22 B23 B21H B23H PIPE21 PIPE21H GK3D2 GK3D2N SAX1}
    set linear_3 {D B31 B32 B33 B34 B32R B31H B33H PIPE31 PIPE31H}
    set triangles_3 {S3 S3R STRI3 STRI35 M3D3 R3D3 DS3}
    set triangles_6 {STRI65 DS6 M3D6 S6 CAX6}
    set quadrilateral_4 {S4 S4R S4R5 M3D4 M3D4R R3D4 DS4 CPS4}
    set quadrilateral_8 {S8R S8R5 S8RT DS8 M3D8 M3D8R S8 CAX8 CAX8R CPS8 CPS8R}
    set quadrilateral_9 {S9}
    set tetrahedra_4 {C3D4 C3D4H DC3D4 C3D4E DC3D4E}
    set tetrahedra_10 {C3D10 C3D10H C3D10M C3D10MH DC3D10 C3D10E DC3D10E}
    set hexahedra_8 {C3D8 C3D8T C3D8H C3D8HT C3D8I C3D8IH C3D8R C3D8RH C3D8E DC3D8 AC3D8 DC3D8E DCC3D8D GK3D8 DK3D8N F3D8}
    set hexahedra_20 {C3D20 C3D20H C3D20R C3D20RH DC3D20 AC3D20 C3D20E CDD20RE C3D20T C3D20HT C3D20RT C3D20RHT DC3D20E}
    set hexahedra_27 {C3D27}
    set prism_6 {C3D6 C3D6H DC3D6 C3D6E DC3D6E GK3D6 GK3D6N F3D6}
    set prism_12 {GK3D12M}
    set prism_15 {C3D15 C3D15H DC3D15 C3D15E DC3D15E}
    if { [lsearch $linear_2 $abaqus_element_type] != -1  } {
        set elem_type Linear
        set element_nnodes 2
    } elseif { [lsearch $linear_3 $abaqus_element_type] != -1  } {
        set elem_type Linear
        set element_nnodes 3
    } elseif { [lsearch $triangles_3 $abaqus_element_type] != -1  } {
        set elem_type Triangle
        set element_nnodes 3
    } elseif { [lsearch $triangles_6 $abaqus_element_type] != -1  } {
        set elem_type Triangle
        set element_nnodes 6
    } elseif { [lsearch $quadrilateral_4 $abaqus_element_type] != -1  } {
        set elem_type Quadrilateral
        set element_nnodes 4
    } elseif { [lsearch $quadrilateral_8 $abaqus_element_type] != -1  } {
        set elem_type Quadrilateral
        set element_nnodes 8
    } elseif { [lsearch $quadrilateral_9 $abaqus_element_type] != -1  } {
        set elem_type Quadrilateral
        set element_nnodes 9
    } elseif { [lsearch $tetrahedra_4 $abaqus_element_type] != -1  } {
        set elem_type Tetrahedra
        set element_nnodes 4
    } elseif { [lsearch $tetrahedra_10 $abaqus_element_type] != -1  } {
        set elem_type Tetrahedra
        set element_nnodes 10
    } elseif { [lsearch $hexahedra_8 $abaqus_element_type] != -1  } {
        set elem_type Hexahedra
        set element_nnodes 8
    } elseif { [lsearch $hexahedra_20 $abaqus_element_type] != -1  } {
        set elem_type Hexahedra
        set element_nnodes 20
    } elseif { [lsearch $hexahedra_27 $abaqus_element_type] != -1  } {
        set elem_type Hexahedra
        set element_nnodes 27
    } elseif { [lsearch $prism_6 $abaqus_element_type] != -1  } {
        set elem_type Prism
        set element_nnodes 6
    } elseif { [lsearch $prism_12 $abaqus_element_type] != -1  } {
        set elem_type Prism
        set element_nnodes 12
    } elseif { [lsearch $prism_15 $abaqus_element_type] != -1  } {
        set elem_type Prism
        set element_nnodes 15
    } else {
        W "Abaqus::GetGiDElementTypeAndNnodes. Unexpected element type $abaqus_element_type"
        set elem_type ""
        set element_nnodes 0                  
    }
    return [list $elem_type $element_nnodes]
}

proc Abaqus::ReadPreMesh { filename } {
    variable debug
    set fail 0
    set keys_to_ignore [lsort -dictionary {AMPLITUDE {BEAM SECTION} BOUNDARY BOUNDARYF BUCKLE CFLUX 
            {CHANGE FRICTION} {CHANGE MATERIAL} {CHANGE PLASTIC} CLOAD CONDUCTIVITY 
            {CONTACT FILE} {CONTACT PAIR} CONTROLS {COUPLED TEMPERATURE-DISPLACEMENT} CREEP {CYCLIC HARDENING} 
            DAMPING {DEFORMATION PLASTICITY} DENSITY DEPVAR DFLUX DLOAD DYNAMIC
            {EL FILE} {EL PRINT} ELASTIC ELSET {END STEP} ENDSTEP EQUATION EXPANSION {FACE PRINT} FILM {FLUID CONSTANTS} {FLUID SECTION} 
            FREQUENCY FRICTION HEADING {HEAT TRANSFER} HYPERELASTIC HYPERFOAM
            {INITIAL CONDITIONS} MATERIAL MEMBRANE {MODAL DAMPING} {MODAL DYNAMIC} {MODEL CHANGE} MPC {NO ANALYSIS} 
            {NODE FILE} {NODE PRINT} ORIENTATION {PHYSICAL CONSTANTS} PLASTIC RADIATE RESTART {RIGID BODY}
            {SHELL SECTION} {SOLID SECTION} {SPECIFIC GAS CONSTANT} {SPECIFIC HEAT} STATIC {STEADY STATE DYNAMICS} STEP
            SURFACE {SURFACE BEHAVIOR} {SURFACE INTERACTION} TEMPERATURE
            TIE {TIME POINTS} TRANSFORM TRANSFORMF {USER MATERIAL} {VALUES AT INFINITY} VIEWFACTOR VISCO}]
    #offset_nodes to be added to the number if a previous model exists
    set offset_nodes [GiD_Info Mesh MaxNumNodes]
    set offset_elements [GiD_Info Mesh MaxNumElements]    
    set layer [file rootname [file tail $filename]]
    if { ![GiD_Layers exists $layer] } {
        GiD_Layers create $layer
    }
    GiD_Layers edit to_use $layer

    set fp [open $filename r]
    if { $fp == "" } {
        WarnWin [_ "Cannot open file '%s'" $filename]
        return 1
    }
    set must_read 1
    while { ![eof $fp] } {
        if { $must_read } {
            gets $fp line
        } else {
            set must_read 1
        }
        if { $line == "" } {
            continue
        } 
        if { [string range $line 0 1] == "**" } {
            #comment
            continue
        }
        if { [string index $line 0] == "*" } {
            set items [split $line ,]            
            set key [string toupper [string trim [string range [lindex $items 0] 1 end]]]
            #extra attributes
            array unset abaqus_key_attributes
            foreach item [lrange $items 1 end] {
                lassign [split $item =] subkey value
                set abaqus_key_attributes([string toupper [string trim $subkey]]) [string trim $value]
            }
            if { $key == "NODE" } {
                #attributes: nset
                while { ![eof $fp] } {
                    gets $fp line
                    if { $line == "" } {
                        continue
                    }
                    if { [string index $line 0] == "*" } {
                        if { [string range $line 0 1] == "**" } {
                            #comment
                            continue
                        } else {
                            set must_read 0
                            break
                        }
                    } else {                                                
                        lassign [string map {, { }} $line] node_id x y z
                        if { $z == "" } {
                            set z 0
                        }
                        if { $offset_nodes } {
                            incr node_id $offset_nodes
                        }                        
                        GiD_Mesh create node $node_id [list $x $y $z]
                    }
                }
                
                
            } elseif { $key == "ELEMENT" } {
                #attributes: type elset 
                lassign [Abaqus::GetGiDElementTypeAndNnodes [string toupper $abaqus_key_attributes(TYPE)]] elem_type element_nnodes                
                while { ![eof $fp] } {
                    gets $fp line
                    if { $line == "" } {
                        continue
                    }
                    if { [string index $line 0] == "*" } {
                        if { [string range $line 0 1] == "**" } {
                            #comment
                            continue
                        } else {
                            #GiD_MeshPre_Create $elem_type $element_nnodes $node_ids $vertices $element_ids $element_vertex_indices $zero_based_array $layer
                            set must_read 0                            
                            break
                        }                        
                    } else {
                        while { [string index $line end] == "," && ![eof $fp] } {                            
                            gets $fp line_extra
                            set line $line$line_extra
                        }
                        set element_items [string map {, { }} $line]
                        set element_id [lindex $element_items 0]
                        set element_connectivities [list]
                        foreach node_id [lrange $element_items 1 end] {
                            if { $offset_nodes } {                            
                                incr node_id $offset_nodes
                            }
                            lappend element_connectivities $node_id
                        }
                        if { $elem_type != "" } {
                            if { $elem_type == "Hexahedra" && $element_nnodes == 20 } {
                                #Abaqus has different connectivities that GiD
                                set block_1 [lrange $element_connectivities 0 11]
                                set block_2 [lrange $element_connectivities 12 15]
                                set block_3 [lrange $element_connectivities 16 19]
                                set element_connectivities [concat $block_1 $block_3 $block_2]
                            } elseif { $elem_type == "Prism" && $element_nnodes == 15 } {
                                #Abaqus has different connectivities that GiD
                                set block_1 [lrange $element_connectivities 0 8]
                                set block_2 [lrange $element_connectivities 9 11]
                                set block_3 [lrange $element_connectivities 12 14]
                                set element_connectivities [concat $block_1 $block_3 $block_2]
                            }
                            if { [catch { GiD_Mesh create element append $elem_type $element_nnodes $element_connectivities } err] } {
                                W $err
                            }
                        }
                    }
                }
            } elseif { $key == "NSET" } {
                #attributes: nset 
                while { ![eof $fp] } {
                    gets $fp line
                    if { $line == "" } {
                        continue
                    }
                    if { [string index $line 0] == "*" } {
                        if { [string range $line 0 1] == "**" } {
                            #comment
                            continue
                        } else {
                            set must_read 0
                            break
                        }                        
                    } else {
                       #do something...
                    }
                }
            } elseif { [lsearch -dictionary -sorted $keys_to_ignore $key] != -1 } {
                #attributes:
                while { ![eof $fp] } {
                    gets $fp line
                    if { [string index $line 0] == "*" } {
                        if { [string range $line 0 1] == "**" } {
                            #comment                            
                            continue
                        } else {
                            set must_read 0
                            break
                        }                        
                    } else {
                       #do something...                        
                    }
                }                
            } elseif { $key == "*" } {
                #comment
            } elseif { $key == "" } {
                #???
            } else {
                set unimplemented_types($key) 1                
            }                        
        } else {
            W "Abaqus::ReadPre. Ignored line $line"            
        }            
    }
    
    close $fp

    if { $offset_nodes != [GiD_Info Mesh MaxNumNodes] } {
        GiD_ModifiedFileFlag set 1
    }
    if { $debug } {
        if { [llength [array names unimplemented_types]] } {
            W "Unimplemented types: [lsort -dictionary [array names unimplemented_types]]"
        }
    }
    return $fail
}

proc Abaqus::ReadPre0 { filenames } {    
    variable percent_done
    variable stop    
    set fail 0         
    set ::Abaqus::stop 0
    set ::Abaqus::percent_done 0
    set percent_max [llength $filenames]
    ::GidUtils::CreateAdvanceBar [_ "Loading file"] [_ "Process"] ::Abaqus::percent_done $percent_max {set ::Abaqus::stop 1}             
    foreach filename $filenames {
        if { $::Abaqus::stop } {
            break
        }        
        if { [file exists $filename] } {
            set res ""
            if { [catch { set local_fail [Abaqus::ReadPreMesh $filename] } err options] } {                          
                #set fail 1
                #W $err
                GidUtils::DeleteAdvanceBar 
                return -options $options $err
            } else {       
                if { $local_fail } {
                    set fail 1
                }
                RaiseEvent_GenericProc GiD_Event_AfterOpenFile $filename ABAQUS_FORMAT $local_fail
            }
            
        }
        incr ::Abaqus::percent_done            
    }           
    GidUtils::DeleteAdvanceBar 
    return $fail
}

proc Abaqus::AddToMenu { } {    
    if { [GidUtils::IsTkDisabled] } {
        return
    }
    if { [GiDMenu::GetOptionIndex Files [list Import "Abaqus..."] PRE] != -1 } {
        return
    }

    #try to insert this menu after the word "Files->Import->XYZ nodes"
    set position [GiDMenu::GetOptionIndex Files [list Import "XYZ nodes"] PRE]
    if { $position == -1 } {
        set position end
    }
    GiDMenu::InsertOption Files [list Import "Abaqus..."] $position PRE Abaqus::ReadPreW "" "" insertafter _
}
 

#invoke this menus changes
Abaqus::AddToMenu
GiDMenu::UpdateMenus

#register the proc to be automatically called when re-creating all menus (e.g. when doing files new)
GiD_RegisterPluginAddedMenuProc Abaqus::AddToMenu
 
#register the proc to be automatically called when dropping a file
GiD_RegisterExtensionProc ".Abaqus .inp" PRE Abaqus::ReadPreSingleFile

